# Generated file - DO NOT EDIT.  Edit config.py.in instead.

prefix="/usr"
datadir="share"
localedir="/usr/share/locale"
pkgdatadir="/usr/share/warpinator"
libdir="lib"
libexecdir="/usr/lib/warpinator"
include_firewall_mod=True
bundle_zeroconf=False
bundle_grpc=False
PACKAGE="warpinator"
VERSION="2.0.3"
GETTEXT_PACKAGE="warpinator"
FLATPAK_BUILD=False
RPC_API_VERSION="2"



###################### This is not a constant #########################
# sandbox_mode is set in warpinator.py __main__ at launch.
sandbox_mode = "legacy"
